@@ -1,93 +1,14 @@
-/*	This is pstree written by Fred Hucht (c) 1993-2002	*
- *	EMail: fred@thp.Uni-Duisburg.DE				*
+/*	This is pstree written by Fred Hucht (c) 1993-2004	*
+ *	EMail: fred AT thp.Uni-Duisburg.de			*
  *	Feel free to copy and redistribute in terms of the	*
  * 	GNU public license. 					*
  *
- * $Id: pstree.c,v 2.17 2001-12-17 13:18:02+01 fred Exp $
- *
- * $Log: pstree.c,v $
- * Revision 2.17  2001-12-17 13:18:02+01  fred
- * Changed ps call to something like ps -eo uid,pid,ppid,pgid,args under
- * AIX and Linux, workaround for AIX 5L.
- *
- * Revision 2.17  2001-12-13 08:27:00+08  chris
- * Added workaround for AIX Version >= 5
- *
- * Revision 2.16  2000-03-01 10:42:22+01  fred
- * Added support for thread count (thcount) in other OSs than AIX
- *
- * Revision 2.15  2000-03-01 10:18:56+01  fred
- * Added process group support for {Net|Open}BSD following a suggestion
- * by Ralf Meyer <ralf@thp.Uni-Duisburg.de>
- *
- * Revision 2.14  1999-03-22 20:45:02+01  fred
- * Fixed bug when line longer than MAXLINE, set MAXLINE=512
- *
- * Revision 2.13  1998-12-17 19:31:53+01  fred
- * Fixed problem with option -f when input file is empty
- *
- * Revision 2.12  1998-12-07 17:08:59+01  fred
- * Added -f option and sun 68000 support by Paul Kern
- * <pkern@utcc.utoronto.ca>
- *
- * Revision 2.11  1998-05-23 13:30:28+02  fred
- * Added vt100 sequences, NetBSD support
- *
- * Revision 2.10  1998-02-02 15:04:57+01  fred
- * Fixed bug in MakeTree()/get_pid_index() when parent doesn't
- * exist. Thanks to Igor Schein <igor@andrew.air-boston.com> for the bug
- * report.
- *
- * Revision 2.9  1998-01-07 16:55:26+01  fred
- * Added support for getprocs()
- *
- * Revision 2.9  1998-01-06 17:13:19+01  fred
- * Added support for getprocs() under AIX
- *
- * Revision 2.8  1997-10-22 15:09:39+02  fred
- * Cosmetic
- *
- * Revision 2.7  1997-10-22 15:01:40+02  fred
- * Minor changes in getprocs for AIX
- *
- * Revision 2.6  1997/10/16 16:35:19  fred
- * Added uid2name() caching username lookup, added patch for Solaris 2.x
- *
- * Revision 2.5  1997-02-05 14:24:53+01  fred
- * return PrintTree when nothing to do.
- *
- * Revision 2.4  1997/02/05 09:54:08  fred
- * Fixed bug when P[i].cmd is empty
- *
- * Revision 2.3  1997-02-04 18:40:54+01  fred
- * Cosmetic
- *
- * Revision 2.2  1997-02-04 14:11:17+01  fred
- * *** empty log message ***
- *
- * Revision 2.1  1997-02-04 13:55:14+01  fred
- * Rewritten
- *
- * Revision 1.13  1997-02-04 09:01:59+01  fred
- * Start of rewrite
- *
- * Revision 1.12  1996-09-17 21:54:05+02  fred
- * *** empty log message ***
- *
- * Revision 1.11  1996-09-17 21:52:52+02  fred
- * revision added
- *
- * Revision 1.10  1996-09-17 21:45:35+02  fred
- * replace \n and \t with ? in output
- *
- * Revision 1.4  1996-09-17 21:43:14+02  fred
- * Moved under RCS, replace \n and \t with ?
+ * $Id: pstree.c,v 2.25 2004-05-14 16:41:39+02 fred Exp $
  */
-
 static char *WhatString[]= {
-  "@(#)pstree $Revision: 2.17 $ by Fred Hucht (C) 1993-2002",
-  "@(#)EMail:fred@thp.Uni-Duisburg.de",
-  "$Id: pstree.c,v 2.17 2001-12-17 13:18:02+01 fred Exp $"
+  "@(#)pstree $Revision: 2.25 $ by Fred Hucht (C) 1993-2004",
+  "@(#)EMail: fred AT thp.Uni-Duisburg.de",
+  "$Id: pstree.c,v 2.25 2004-05-14 16:41:39+02 fred Exp $"
 };
 
 #define MAXLINE 512
@@ -96,13 +17,12 @@ static char *WhatString[]= {
 /* Under AIX, we directly read the process table from the kernel */
 # ifndef _AIX50
 /* problems with getprocs() under AIX 5L
- * workaround contributed by Chris Benesch <chris@fdbs.com> */
+ * workaround contributed by Chris Benesch <chris AT fdbs.com> */
 #  define USE_GetProcessesDirect
 # endif /*_AIX50*/
 #  define HAS_TERMDEF
 extern char *termdef(int, char);
 #  define _ALL_SOURCE
-/*#  include <pwd.h>*/
 #  include <procinfo.h>
 #  define USE_GETPROCS
 
@@ -126,26 +46,44 @@ extern getargs(struct ProcInfo *, int, char *, int);
 #  define PSCMD 	"ps -eko uid,pid,ppid,pgid,thcount,args"
 #  define PSFORMAT 	"%ld %ld %ld %ld %ld %[^\n]"
 #  define PSVARS	&P[i].uid, &P[i].pid, &P[i].ppid, &P[i].pgid, &P[i].thcount, P[i].cmd
-
-#elif defined(__linux)	/* Linux */
-#  define USE_GetProcessesDirect
-#  define HAS_PGID
-#  include <glob.h>
-#  include <sys/stat.h>
+#  define PSVARSN	6
+/************************************************************************/
+#elif defined(__linux) || (defined __alpha && defined(_SYSTYPE_BSD))
+/* TRU64 contributed by Frank Parkin <fparki AT acxiom.co.uk>
+ */
+#  ifdef __linux
+#    define USE_GetProcessesDirect
+#    include <glob.h>
+#    include <sys/stat.h>
+#  endif
 #  define UID2USER
-/*#  define PSCMD 	"ps laxw"
-  #  define PSFORMAT 	"%*d %ld %ld %ld %*37c %*s %[^\n]"*/
+#  define HAS_PGID
 #  define PSCMD 	"ps -eo uid,pid,ppid,pgid,args"
 #  define PSFORMAT 	"%ld %ld %ld %ld %[^\n]"
 #  define PSVARS	&P[i].uid, &P[i].pid, &P[i].ppid, &P[i].pgid, P[i].cmd
-/* #elif defined(sparc)	SunOS */
+#  define PSVARSN	5
+/************************************************************************/
+#elif defined(__FreeBSD__) || defined(__NetBSD__) || defined(__OpenBSD__) || defined(__APPLE__)
+/* NetBSD contributed by Gary D. Duzan <gary AT wheel.tiac.net>
+ * FreeBSD contributed by Randall Hopper <rhh AT ct.picker.com> 
+ * Darwin / Mac OS X patch by Yuji Yamano <yyamano AT kt.rim.or.jp>
+ * wide output format fix for NetBSD by Jeff Brown <jabrown AT caida.org>
+ * (Net|Open|Free)BSD & Darwin merged by Ralf Meyer <ralf AT thp.Uni-Duisburg.DE>
+ */
+#  define HAS_PGID
+#  define PSCMD 	"ps -axwwo user,pid,ppid,pgid,command"
+#  define PSFORMAT 	"%s %ld %ld %ld %[^\n]"
+#  define PSVARS	P[i].name, &P[i].pid, &P[i].ppid, &P[i].pgid, P[i].cmd
+#  define PSVARSN	5
+#  define ZOMBIES_HAVE_PID_0
+/************************************************************************/
 #elif defined(sun) && (!defined(__SVR4)) /* Solaris 1.x */
-/* contributed by L. Mark Larsen <mlarsen@ptdcs2.intel.com> */
-/* new cpp criteria by Pierre Belanger <belanger@risq.qc.ca> */
+/* contributed by L. Mark Larsen <mlarsen AT ptdcs2.intel.com> */
+/* new cpp criteria by Pierre Belanger <belanger AT risq.qc.ca> */
 #  define solaris1x
 #  define UID2USER
 #  ifdef mc68000
-/* contributed by Paul Kern <pkern@utcc.utoronto.ca> */
+/* contributed by Paul Kern <pkern AT utcc.utoronto.ca> */
 #    define PSCMD 	"ps laxw"
 #    define PSFORMAT 	"%*7c%ld %ld %ld %*d %*d %*d %*x %*d %*d %*x %*14c %[^\n]"
 #    define uid_t	int
@@ -154,39 +92,36 @@ extern getargs(struct ProcInfo *, int, char *, int);
 #    define PSCMD 	"ps jaxw"
 #    define PSFORMAT 	"%ld %ld %*d %*d %*s %*d %*s %ld %*s %[^\n]"
 #    define PSVARS 	&P[i].ppid, &P[i].pid, &P[i].uid, P[i].cmd
+#    define PSVARSN	4
 #  endif
+/************************************************************************/
 #elif defined(sun) && (defined(__SVR4)) /* Solaris 2.x */
-/* contributed by Pierre Belanger <belanger@risq.qc.ca> */
+/* contributed by Pierre Belanger <belanger AT risq.qc.ca> */
 #  define solaris2x
 #  define PSCMD         "ps -ef"
 #  define PSFORMAT      "%s %ld %ld %*d %*s %*s %*s %[^\n]"
+/************************************************************************/
 #elif defined(bsdi)
-/* contributed by Dean Gaudet <dgaudet@hotwired.com> */
+/* contributed by Dean Gaudet <dgaudet AT hotwired.com> */
 #  define UID2USER
 #  define PSCMD 	"ps laxw"
 #  define PSFORMAT 	"%ld %ld %ld %*d %*d %*d %*d %*d %*s %*s %*s %*s %[^\n]"
-#elif defined(__FreeBSD__) /* FreeBSD  */
-/* FreeBSD contributed by Randall Hopper <rhh@ct.picker.com> */
-#  define PSCMD 	"ps -axo \"user pid ppid command\""
-#  define PSFORMAT 	"%s %d %d %[^\n]"
+/************************************************************************/
 #elif defined(_BSD)	/* Untested */
 #  define UID2USER
 #  define PSCMD 	"ps laxw"
 #  define PSFORMAT 	"%*d %*c %ld %ld %ld %*d %*d %*d %*x %*d %d %*15c %*s %[^\n]"
+/************************************************************************/
 #elif defined(__convex)	/* ConvexOS */
 #  define UID2USER
 #  define PSCMD 	"ps laxw"
 #  define PSFORMAT 	"%*s %ld %ld %ld %*d %*g %*d %*d %*21c %*s %[^\n]"
-#elif defined(__NetBSD__) || defined(__OpenBSD__)
-/* NetBSD contributed by Gary D. Duzan <gary@wheel.tiac.net> */
-#  define HAS_PGID
-#  define PSCMD 	"ps -axwwo user,pid,ppid,pgid,command"
-#  define PSFORMAT 	"%s %ld %ld %ld %[^\n]"
-#  define PSVARS	P[i].name, &P[i].pid, &P[i].ppid, &P[i].pgid, P[i].cmd
+/************************************************************************/
 #else			/* HP-UX, A/UX etc. */
 #  define PSCMD 	"ps -ef"
 #  define PSFORMAT 	"%s %ld %ld %*20c %*s %[^\n]"
 #endif
+/*********************** end of configurable part ***********************/
 
 #ifndef PSVARS 		/* Set default */
 # ifdef UID2USER
@@ -194,6 +129,7 @@ extern getargs(struct ProcInfo *, int, char *, int);
 # else
 #  define PSVARS	P[i].name, &P[i].pid, &P[i].ppid, P[i].cmd
 # endif
+# define PSVARSN	4
 #endif
 
 #include <stdio.h>
@@ -202,6 +138,10 @@ extern getargs(struct ProcInfo *, int, char *, int);
 #include <unistd.h>		/* For getopt() */
 #include <pwd.h>		/* For getpwnam() */
 
+#ifdef DEBUG
+# include <errno.h>
+#endif
+
 #ifdef NEED_STRSTR
 static char *strstr(char *, char *);
 #endif
@@ -212,28 +152,37 @@ static char *strstr(char *, char *);
 #endif
 
 struct TreeChars {
-  char *s2, 		/* String between header and pid */
-    *p, 		/* dito, when parent of printed childs */
-    *pgl,		/* Process group leader */
-    *npgl,		/* No process group leader */
-    *barc, 		/* bar for line with child */
-    *bar, 		/* bar for line without child */
-    *barl,		/* bar for last child */
-    *sg,		/* Start graphics (alt char set) */
-    *eg,		/* End graphics (alt char set) */
-    *init;		/* Init string sent at the beginning */
+  char *s2, 		/* SS String between header and pid */
+    *p, 		/* PP dito, when parent of printed childs */
+    *pgl,		/* G  Process group leader */
+    *npgl,		/* N  No process group leader */
+    *barc, 		/* C  bar for line with child */
+    *bar, 		/* B  bar for line without child */
+    *barl,		/* L  bar for last child */
+    *sg,		/*    Start graphics (alt char set) */
+    *eg,		/*    End graphics (alt char set) */
+    *init;		/*    Init string sent at the beginning */
 };
 
+/* Example:
+ * |-+- 01111 ...        CPPN 01111 ...
+ * | \-+=   01112 ...    B LPPG 01112 ...
+ * |   |--= 01113 ...    B   CSSG 01113 ...
+ * |   \--= 01114 ...    B   LSSG 01114 ...
+ * \-+- 01115 ...        LSSN 01115 ...
+ */
+
 enum { G_ASCII = 0, G_PC850 = 1, G_VT100 = 2, G_LAST };
 
-/* VT sequences contributed by Randall Hopper <rhh@ct.picker.com> */
+/* VT sequences contributed by Randall Hopper <rhh AT ct.picker.com> */
 static struct TreeChars TreeChars[] = {
+  /* SS          PP          G       N       C       B       L      sg      eg      init */
   { "--",       "-+",       "=",    "-",    "|",    "|",    "\\",   "",     "",     ""             }, /*Ascii*/
   { "\304\304", "\304\302", "\372", "\304", "\303", "\263", "\300", "",     "",     ""             }, /*Pc850*/
   { "qq",       "qw",       "`",    "q",    "t",    "x",    "m",    "\016", "\017", "\033(B\033)0" }  /*Vt100*/
 }, *C;
 
-int MyPid, NProc, Columns;
+int MyPid, NProc, Columns, RootPid;
 short showall = TRUE, soption = FALSE, Uoption = FALSE;
 char *name = "", *str = NULL, *Progname;
 long ipid = -1;
@@ -403,7 +352,7 @@ int GetProcessesDirect(void) {
 #ifdef __linux
 int GetProcessesDirect(void) {
   glob_t globbuf;
-  int i;
+  int i, j;
   
   glob("/proc/[0-9]*", GLOB_NOSORT, NULL, &globbuf);
   
@@ -413,54 +362,55 @@ int GetProcessesDirect(void) {
     exit(1);
   }
   
-  for (i = 0; i < globbuf.gl_pathc; i++) {
-    char name[32], line[MAXLINE], c;
+  for (i = j = 0; i < globbuf.gl_pathc; i++) {
+    char name[32], c;
     FILE *tn;
     struct stat stat;
     int k = 0;
     
-    sprintf(name, "%s%s",
-	    globbuf.gl_pathv[globbuf.gl_pathc - i - 1], "/stat");
+    snprintf(name, sizeof(name), "%s%s",
+	     globbuf.gl_pathv[globbuf.gl_pathc - i - 1], "/stat");
     tn = fopen(name, "r");
-    fscanf(tn, "%d %s %*c %d %d",
-	   &P[i].pid, P[i].cmd, &P[i].ppid, &P[i].pgid);
+    if (tn == NULL) continue; /* process vanished since glob() */
+    fscanf(tn, "%ld %s %*c %ld %ld",
+	   &P[j].pid, P[j].cmd, &P[j].ppid, &P[j].pgid);
     fstat(fileno(tn), &stat);
-    P[i].uid = stat.st_uid;
+    P[j].uid = stat.st_uid;
     fclose(tn);
+    P[j].thcount = 1;
     
-    sprintf(name, "%s%s",
-	    globbuf.gl_pathv[globbuf.gl_pathc - i - 1], "/cmdline");
+    snprintf(name, sizeof(name), "%s%s",
+	     globbuf.gl_pathv[globbuf.gl_pathc - i - 1], "/cmdline");
     tn = fopen(name, "r");
+    if (tn == NULL) continue;
     while (k < MAXLINE - 1 && EOF != (c = fgetc(tn))) {
-      P[i].cmd[k++] = c == '\0' ? ' ' : c;
+      P[j].cmd[k++] = c == '\0' ? ' ' : c;
     }
-    if (k > 0) P[i].cmd[k] = '\0';
+    if (k > 0) P[j].cmd[k] = '\0';
     fclose(tn);
     
-    uid2user(P[i].uid, P[i].name, sizeof(P[i].name));
+    uid2user(P[j].uid, P[j].name, sizeof(P[j].name));
     
 #ifdef DEBUG
     if (debug) fprintf(stderr,
-		       "uid=%5ld, name=%8s, pid=%5ld, ppid=%5ld, pgid=%5ld, thcount=%d, cmd='%s'\n",
-		       P[i].uid, P[i].name, P[i].pid, P[i].ppid, P[i].pgid, P[i].thcount, P[i].cmd);
+		       "uid=%5ld, name=%8s, pid=%5ld, ppid=%5ld, pgid=%5ld, thcount=%ld, cmd='%s'\n",
+		       P[j].uid, P[j].name, P[j].pid, P[j].ppid, P[j].pgid, P[j].thcount, P[j].cmd);
 #endif
-    P[i].parent = P[i].child = P[i].sister = -1;
-    P[i].print  = FALSE;
+    P[j].parent = P[j].child = P[j].sister = -1;
+    P[j].print  = FALSE;
+    j++;
   }
   globfree(&globbuf);
-  return i;
+  return j;
 }
 #endif /* __linux */
 
 int GetProcesses(void) {
   FILE *tn;
-  int len, i = 0;  extern int errno; /* For popen() */
-#ifdef UID2USER
-  int xx; /* For testing... */
-#endif
+  int i = 0;
   char line[MAXLINE], command[] = PSCMD;
   
-  /* file read code contributed by Paul Kern <pkern@utcc.utoronto.ca> */
+  /* file read code contributed by Paul Kern <pkern AT utcc.utoronto.ca> */
   if (input != NULL) {
     if (strcmp(input, "-") == 0)
       tn = stdin;
@@ -496,7 +446,9 @@ int GetProcesses(void) {
     exit(1);
   }
   
-  while (NULL != fgets(line, MAXLINE, tn) && 10 < (len = strlen(line))) {
+  while (NULL != fgets(line, MAXLINE, tn)) {
+    int len, num;
+    len = strlen(line);
 #ifdef DEBUG
     if (debug) {
       fprintf(stderr, "len=%3d ", len);
@@ -521,16 +473,23 @@ int GetProcesses(void) {
     { /* SunOS allows columns to run together.  With the -j option, the CPU
        * time used can run into the numeric user id, so make sure there is
        * space between these two columns.  Also, the order of the desired
-       * items is different. (L. Mark Larsen <mlarsen@ptdcs2.intel.com>)
+       * items is different. (L. Mark Larsen <mlarsen AT ptdcs2.intel.com>)
        */
       char buf1[45], buf2[MAXLINE];
       buf1[44] = '\0';
       sscanf(line, "%44c%[^\n]", buf1, buf2);
-      sprintf(line, "%s %s", buf1, buf2);
+      snprintf(line, sizeof(line), "%s %s", buf1, buf2);
     }
 #endif
     
-    sscanf(line, PSFORMAT, PSVARS);
+    num = sscanf(line, PSFORMAT, PSVARS);
+    
+    if (num != PSVARSN) {
+#ifdef DEBUG
+      if (debug) fprintf(stderr, "dropped line, num=%d != %d\n", num, PSVARSN);
+#endif
+      continue;
+    }
     
 #ifdef UID2USER	/* get username */
     uid2user(P[i].uid, P[i].name, sizeof(P[i].name));
@@ -538,7 +497,7 @@ int GetProcesses(void) {
 
 #ifdef DEBUG
     if (debug) fprintf(stderr,
-		      "uid=%5ld, name=%8s, pid=%5ld, ppid=%5ld, pgid=%5ld, thcount=%d, cmd='%s'\n",
+		      "uid=%5ld, name=%8s, pid=%5ld, ppid=%5ld, pgid=%5ld, thcount=%ld, cmd='%s'\n",
 		      P[i].uid, P[i].name, P[i].pid, P[i].ppid, P[i].pgid, P[i].thcount, P[i].cmd);
 #endif
     P[i].parent = P[i].child = P[i].sister = -1;
@@ -552,10 +511,49 @@ int GetProcesses(void) {
   return i;
 }
 
+int GetRootPid(void) {
+  int me;
+  for (me = 0; me < NProc; me++) {
+    if (P[me].pid == -1) return P[me].pid;
+  }
+  /* PID == 1 not found, so we'll take process with PPID == 0
+   * Fix for TRU64 TruCluster with uniq PIDs
+   * reported by Frank Parkin <fparki AT acxiom.co.uk>
+   * re-reported by Eric van Doorn <Eric.van.Doorn AT isc.politie.nl>,
+   * because fix was not published by me :-/ */
+  for (me = 0; me < NProc; me++) {
+    if (P[me].ppid == 0) return P[me].pid;
+  }
+  /* Should not happen */
+  fprintf(stderr,
+	  "%s: No process found with PID == 1 || PPID == 0, contact author.\n",
+	  Progname);
+  exit(1);
+}
+
+#ifdef ZOMBIES_HAVE_PID_0
+void FixZombies(void) {
+  int me, num = 0;
+  for (me = 0; me < NProc; me++) {
+    if (P[me].pid == 0) num++;
+  }
+  if (num > 1) for (me = 0; me < NProc; me++) {
+    if (P[me].pid == 0 && P[me].ppid != 0 && P[me].ppid != -1) {
+      P[me].pid = -1;
+#ifdef DEBUG
+      if (debug) fprintf(stderr,
+			 "fixed zombie %s with ppid %d\n",
+			 P[me].cmd, P[me].ppid);
+#endif
+    }
+  }
+}
+#endif
+
 int get_pid_index(long pid) {
-  int i;
-  for (i = NProc - 1;i >= 0 && P[i].pid != pid; i--); /* Search process */
-  return i;
+  int me;
+  for (me = NProc - 1;me >= 0 && P[me].pid != pid; me--); /* Search process */
+  return me;
 }
 
 #define EXIST(idx) ((idx) != -1)
@@ -563,67 +561,86 @@ int get_pid_index(long pid) {
 void MakeTree(void) {
   /* Build the process hierarchy. Every process marks itself as first child
    * of it's parent or as sister of first child of it's parent */
-  int i, idx;
-  
-  for (i = 0; i < NProc; i++) {
-    idx = get_pid_index(P[i].ppid);
-    if (idx != i && idx != -1) { /* valid process, not me */
-      P[i].parent = idx;
-      if (P[idx].child == -1)
-	P[idx].child = i;
+  int me;  
+  for (me = 0; me < NProc; me++) {
+    int parent;
+    parent = get_pid_index(P[me].ppid);
+    if (parent != me && parent != -1) { /* valid process, not me */
+      P[me].parent = parent;
+      if (P[parent].child == -1) /* first child */
+	P[parent].child = me;
       else {
-	for (idx = P[idx].child; EXIST(P[idx].sister); idx = P[idx].sister);
-	P[idx].sister = i;
+	int sister;
+	for (sister = P[parent].child; EXIST(P[sister].sister); sister = P[sister].sister);
+	P[sister].sister = me;
       }
     }
   }
 }
 
-void MarkChildren(int i) {
-  int idx;
-  P[i].print = TRUE;
-  for (idx = P[i].child; EXIST(idx); idx = P[idx].sister)
-    MarkChildren(idx);
+void MarkChildren(int me) {
+  int child;
+  P[me].print = TRUE;
+  for (child = P[me].child; EXIST(child); child = P[child].sister)
+    MarkChildren(child);
 }
 
 void MarkProcs(void) {
-  int i;
-  for (i = 0; i < NProc; i++) {
+  int me;
+  for (me = 0; me < NProc; me++) {
     if (showall) {
-      P[i].print = TRUE;
+      P[me].print = TRUE;
     } else {
       int parent;
-      if (0 == strcmp(P[i].name, name) 		/* for -u */
+      if (0 == strcmp(P[me].name, name)		/* for -u */
 	 || (Uoption &&
-	     0 != strcmp(P[i].name, "root"))	/* for -U */
-	 || P[i].pid == ipid			/* for -p */
+	     0 != strcmp(P[me].name, "root"))	/* for -U */
+	 || P[me].pid == ipid			/* for -p */
 	 || (soption
-	     && NULL != strstr(P[i].cmd, str)
-	     && P[i].pid != MyPid)		/* for -s */
+	     && NULL != strstr(P[me].cmd, str)
+	     && P[me].pid != MyPid)		/* for -s */
 	 ) {
 	/* Mark parents */
-	for (parent = P[i].parent; EXIST(parent); parent = P[parent].parent) {
+	for (parent = P[me].parent; EXIST(parent); parent = P[parent].parent) {
 	  P[parent].print = TRUE;
 	}
 	/* Mark children */
-	MarkChildren(i);
+	MarkChildren(me);
       }
     }
+#if 0 /* experimental thread compression */
+    {
+      int parent = P[me].parent;
+      int ancestor; /* oldest parent with same cmd */
+      if (0 == strcmp(P[me].cmd, P[parent].cmd)) {
+	P[me].print = FALSE;
+	for (parent = P[me].parent;
+	     EXIST(parent) && (0 == strcmp(P[me].cmd, P[parent].cmd));
+	     parent = P[parent].parent) {
+	  ancestor = parent;
+	}
+	fprintf(stderr, "%d: %d\n",
+		P[me].pid,
+		P[ancestor].pid);
+	P[ancestor].thcount++;
+      }
+    }
+#endif
   }
 }
 
 void DropProcs(void) {
-  int i;
-  for (i = 0; i < NProc; i++) if (P[i].print) {
-    int idx;
+  int me;
+  for (me = 0; me < NProc; me++) if (P[me].print) {
+    int child, sister;
     /* Drop children that won't print */
-    for (idx = P[i].child;
-	EXIST(idx) && !P[idx].print; idx = P[idx].sister);
-    P[i].child = idx;
+    for (child = P[me].child;
+	 EXIST(child) && !P[child].print; child = P[child].sister);
+    P[me].child = child;
     /* Drop sisters that won't print */
-    for (idx = P[i].sister;
-	EXIST(idx) && !P[idx].print; idx = P[idx].sister);
-    P[i].sister = idx;
+    for (sister = P[me].sister;
+	 EXIST(sister) && !P[sister].print; sister = P[sister].sister);
+    P[me].sister = sister;
   }
 }
 
@@ -633,27 +650,27 @@ void PrintTree(int idx, const char *head) {
   
   if (head[0] == '\0' && !P[idx].print) return;
   
-  if (P[idx].thcount > 1) sprintf(thread, "[%ld]", P[idx].thcount);
+  if (P[idx].thcount > 1) snprintf(thread, sizeof(thread), "[%ld]", P[idx].thcount);
   
-  sprintf(out,
-	  "%s%s%s%s%s%s %05ld %s %s%s" /*" (ch=%d, si=%d, pr=%d)"*/,
-	  C->sg,
-	  head,
-	  head[0] == '\0' ? "" : EXIST(P[idx].sister) ? C->barc : C->barl,
-	  EXIST(P[idx].child)       ? C->p   : C->s2,
-	  P[idx].pid == P[idx].pgid ? C->pgl : C->npgl,
-	  C->eg,
-	  P[idx].pid, P[idx].name,
-	  thread,
-	  P[idx].cmd
-	  /*,P[idx].child,P[idx].sister,P[idx].print*/);
+  snprintf(out, sizeof(out),
+	   "%s%s%s%s%s%s %05ld %s %s%s" /*" (ch=%d, si=%d, pr=%d)"*/,
+	   C->sg,
+	   head,
+	   head[0] == '\0' ? "" : EXIST(P[idx].sister) ? C->barc : C->barl,
+	   EXIST(P[idx].child)       ? C->p   : C->s2,
+	   P[idx].pid == P[idx].pgid ? C->pgl : C->npgl,
+	   C->eg,
+	   P[idx].pid, P[idx].name,
+	   thread,
+	   P[idx].cmd
+	   /*,P[idx].child,P[idx].sister,P[idx].print*/);
   
   out[Columns-1] = '\0';
   puts(out);
   
   /* Process children */
-  sprintf(nhead, "%s%s ", head,
-	  head[0] == '\0' ? "" : EXIST(P[idx].sister) ? C->bar : " ");
+  snprintf(nhead, sizeof(nhead), "%s%s ", head,
+	   head[0] == '\0' ? "" : EXIST(P[idx].sister) ? C->bar : " ");
   
   for (child = P[idx].child; EXIST(child); child = P[child].sister)
     PrintTree(child, nhead);
@@ -766,11 +783,25 @@ int main(int argc, char **argv) {
   NProc = GetProcesses();
 #endif
   
+#ifdef ZOMBIES_HAVE_PID_0
+  FixZombies();
+#endif
+  
   if (NProc == 0) {
-    fprintf(stderr, "No processes read.\n");
+    fprintf(stderr, "%s: No processes read.\n", Progname);
     exit(1);
   }
+
+#ifdef DEBUG
+  if (debug) fprintf(stderr, "NProc = %d processes found.\n", NProc);
+#endif
   
+  RootPid = GetRootPid();
+
+#ifdef DEBUG
+  if (debug) fprintf(stderr, "RootPid = %d.\n", RootPid);
+#endif
+
 #if defined(UID2USER) && defined(DEBUG)
   if (debug) uid2user(0,NULL,0);
 #endif
@@ -782,21 +813,24 @@ int main(int argc, char **argv) {
 #ifdef HAS_TERMDEF
     Columns = atoi((char*)termdef(fileno(stdout),'c'));
 #else
-    Columns = 80;
+    char *env = getenv("COLUMNS");
+    Columns = env ? atoi(env) : 80;
 #endif
   }
-  if (Columns == 0 || Columns >= MAXLINE) Columns = MAXLINE - 1;
+  if (Columns == 0) Columns = MAXLINE - 1;
   
   printf("%s", C->init);
   
   Columns += strlen(C->sg) + strlen(C->eg); /* Don't count hidden chars */
+
+  if (Columns >= MAXLINE) Columns = MAXLINE - 1;
   
   MakeTree();
   MarkProcs();
   DropProcs();
   
   if (argc == optind) { /* No pids */
-    PrintTree(get_pid_index(1), "");
+    PrintTree(get_pid_index(RootPid), "");
   } else while (optind < argc) {
     int idx;
     pid = (long)atoi(argv[optind]);
@@ -809,7 +843,7 @@ int main(int argc, char **argv) {
 }
 
 #ifdef NEED_STRSTR
-/* Contributed by Paul Kern <pkern@utcc.utoronto.ca> */
+/* Contributed by Paul Kern <pkern AT utcc.utoronto.ca> */
 static char * strstr(s1, s2)
      register char *s1, *s2;
 {
@@ -822,3 +856,109 @@ static char * strstr(s1, s2)
   return NULL;
 }
 #endif /* NEED_STRSTR */
+
+/*
+ * $Log: pstree.c,v $
+ * Revision 2.25  2004-05-14 16:41:39+02  fred
+ * Added workaround for spurious blank lines in ps output under AIX 5.2
+ * reported by Dean Rowswell <rowswell AT ca.ibm.com>
+ *
+ * Revision 2.24  2004-04-14 09:10:29+02  fred
+ * *** empty log message ***
+ *
+ * Revision 2.23  2004-02-16 10:55:20+01  fred
+ * Fix for zombies (pid == 0) under FreeBSD
+ *
+ * Revision 2.22  2003-12-12 10:58:46+01  fred
+ * Added support for TRU64 v5.1b TruCluster
+ *
+ * Revision 2.21  2003-10-06 13:55:47+02  fred
+ * Fixed SEGV under Linux when process table changes during run
+ *
+ * Revision 2.20  2003-07-09 20:07:29+02  fred
+ * cosmetic
+ *
+ * Revision 2.19  2003/05/26 15:33:35  fred
+ * Merged FreeBSD, (Open|Net)BSD; added Darwin (APPLE), fixed wide output
+ * in FreeBSD
+ *
+ * Revision 2.18  2003/03/13 18:53:22  fred
+ * Added getenv("COLUMNS"), cosmetic changes
+ *
+ * Revision 2.17  2001/12/17 12:18:02  fred
+ * Changed ps call to something like ps -eo uid,pid,ppid,pgid,args under
+ * AIX and Linux, workaround for AIX 5L.
+ *
+ * Revision 2.17  2001-12-13 08:27:00+08  chris
+ * Added workaround for AIX Version >= 5
+ *
+ * Revision 2.16  2000-03-01 10:42:22+01  fred
+ * Added support for thread count (thcount) in other OSs than AIX
+ *
+ * Revision 2.15  2000-03-01 10:18:56+01  fred
+ * Added process group support for {Net|Open}BSD following a suggestion
+ * by Ralf Meyer <ralf AT thp.Uni-Duisburg.de>
+ *
+ * Revision 2.14  1999-03-22 20:45:02+01  fred
+ * Fixed bug when line longer than MAXLINE, set MAXLINE=512
+ *
+ * Revision 2.13  1998-12-17 19:31:53+01  fred
+ * Fixed problem with option -f when input file is empty
+ *
+ * Revision 2.12  1998-12-07 17:08:59+01  fred
+ * Added -f option and sun 68000 support by Paul Kern
+ * <pkern AT utcc.utoronto.ca>
+ *
+ * Revision 2.11  1998-05-23 13:30:28+02  fred
+ * Added vt100 sequences, NetBSD support
+ *
+ * Revision 2.10  1998-02-02 15:04:57+01  fred
+ * Fixed bug in MakeTree()/get_pid_index() when parent doesn't
+ * exist. Thanks to Igor Schein <igor AT andrew.air-boston.com> for the bug
+ * report.
+ *
+ * Revision 2.9  1998-01-07 16:55:26+01  fred
+ * Added support for getprocs()
+ *
+ * Revision 2.9  1998-01-06 17:13:19+01  fred
+ * Added support for getprocs() under AIX
+ *
+ * Revision 2.8  1997-10-22 15:09:39+02  fred
+ * Cosmetic
+ *
+ * Revision 2.7  1997-10-22 15:01:40+02  fred
+ * Minor changes in getprocs for AIX
+ *
+ * Revision 2.6  1997/10/16 16:35:19  fred
+ * Added uid2name() caching username lookup, added patch for Solaris 2.x
+ *
+ * Revision 2.5  1997-02-05 14:24:53+01  fred
+ * return PrintTree when nothing to do.
+ *
+ * Revision 2.4  1997/02/05 09:54:08  fred
+ * Fixed bug when P[i].cmd is empty
+ *
+ * Revision 2.3  1997-02-04 18:40:54+01  fred
+ * Cosmetic
+ *
+ * Revision 2.2  1997-02-04 14:11:17+01  fred
+ * *** empty log message ***
+ *
+ * Revision 2.1  1997-02-04 13:55:14+01  fred
+ * Rewritten
+ *
+ * Revision 1.13  1997-02-04 09:01:59+01  fred
+ * Start of rewrite
+ *
+ * Revision 1.12  1996-09-17 21:54:05+02  fred
+ * *** empty log message ***
+ *
+ * Revision 1.11  1996-09-17 21:52:52+02  fred
+ * revision added
+ *
+ * Revision 1.10  1996-09-17 21:45:35+02  fred
+ * replace \n and \t with ? in output
+ *
+ * Revision 1.4  1996-09-17 21:43:14+02  fred
+ * Moved under RCS, replace \n and \t with ?
+ */
